#!/usr/bin/perl
# Serge Sharoff, University of Leeds
# This is a simple wrapper to Jan Pomikalek's Text::DeDuper;

use strict;
use warnings;
use Getopt::Long;
use Text::DeDuper;
#use Text::PotaModule; #once it's on CPAN
use PotaModule;


my (%params);
my $recordseparator='</text>';
my ($logfile,$tokenise,$iddetector,$glue,$stoplist,$help);

Getopt::Long::Configure('auto_help');
GetOptions(
	   "encoding=s" => \$params{'encoding'},
	   "glue" => \$glue,
	   "help" => \$help,
	   "iddetector=s" => \$iddetector,
	   "journalfile=s" => \$logfile,
	   "ngram_size=i" => \$params{'ngram_size'},
	   "recordseparator=s" => \$recordseparator,
	   "stoplist=s" => \$stoplist,
	   "tokenise" => \$tokenise,
	   "treshold=i" => \$params{'sim_trsh'},
);

if ($help) {
    die "The script finds duplicates in a collection of documents.  For more information run\n  perldoc $0\n";
}

if ($logfile) {
    open(LOG,">$logfile") or die "Cannot create logfile $logfile: $!\n";
} else {
    *LOG=*STDERR;
};

my $deduper = new Text::DeDuper(%params);

if ($stoplist) {
    if ($stoplist=~/ /) { # a list
	$deduper->stoplist(split /\s+/, $stoplist);
    } else { # a file
	$deduper->stoplist($stoplist);
    }
}

$/=$recordseparator;
my $id=0;
while (<>) {
    next unless /\S/;
    if ($tokenise) {
	$_=join ' ',character_tokeniser($_);
    };
    if (my @simlist=$deduper->find_similar($_)) {
	my ($textid)= (/^\s*<text id="(.+?)"/) ? $1 : $id;
	print LOG "--Deleted $textid as similar to @simlist\n";
    } else {
	$id++;
	$deduper->add_doc("$id", $_);
	print LOG "keeping $id\n";
	if ($glue) {
	    glueprint($_);
	} else {
	    print;
	}
    }
};

=head1 NAME
I<deduper.pl>

=head1 SYNOPSIS

cat en-*.txt | deduper.pl -e latin1 -l english-dup.log -s "the is are of and to do" >en-corpus.txt

cat zh-*.txt | deduper.pl -e UTF8 -c -g -t 0.4 -l chinese-dup.log >zh-corpus.txt

=head1 DESCRIPTION

I<deduper.pl>: a simple filter script that reads the input stream and outputs
only documents that are not duplicates or near-duplicates of other
documents in the collection.

=head2 ARGUMENT AND OPTIONS

=item -c --cjk	Tokenise East Asian scripts by adding a space after each non-Latin character 

=item -e --encoding S	Sets the encoding (deafult utf8)

=item -g --glue	Remove whitespace produced by splitting East Asian scripts

=item -h --help	Show a help page and quits

=item -j --journalfile S	File name for writing logs (default STDERR)

=item -n --ngram_size I		Number of words per ngram (default 5)

=item -r --recordseparator S	The separator of documents in the input flow (default </text>)

=item -s --stoplist S		Stop words not to be included in ngrams; either a file name or white-spaced list

=item -t --treshold I	The threshold for the proportion of shared n-grams. 0 means that documents share no n-grams (default 0.2)


=head1 DEPENDENCIES

You need to have the following modules installed:

use Text::DeDuper;

use Text::PotaModule;

To install a module run (as the superuser)

perl -MCPAN -e shell

install module-name


=head1 AUTHOR

Serge Sharoff, University of Leeds

=head1 ACKNOWLEDGMENTS

The script is a simple wrapper around Jan Pomikalek's Text::DeDuper module.

=head1 BUGS

Probably many: if you find one, please let me know.

=head1 COPYRIGHT

Copyright 2006, Serge Sharoff

This program is free software. You may copy or redistribute it under
the same terms as Perl itself.

=head1 SEE ALSO

The PotaModule and other software from the WaCky project:

http://sslmitdev-online.sslmit.unibo.it/wac/wac.php

=cut
    
